<?php
/**
* Community Builder (TM)
* @version $Id: $
* @package CommunityBuilder
* @copyright (C) 2004-2022 www.joomlapolis.com / Lightning MultiCom SA - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

namespace CB\Plugin\AntiSpam;

use CBLib\Application\Application;
use CBLib\Language\CBTxt;
use CBLib\Registry\Registry;

\defined('CBLIB') or die();

class Captcha
{
	/** @var Registry $params */
	protected Registry $params;
	/** @var string $name */
	protected string $name				=	'';
	/** @var string $mode */
	protected string $mode				=	'internal';
	/** @var string $code */
	protected string $code				=	'';
	/** @var string $answer */
	protected string $answer			=	'';
	/** @var string $error */
	protected string $error				=	'';
	/** @var array $input */
	protected array $input				=	[];

	/**
	 * Constructor for captcha object
	 *
	 * @param string $name
	 * @param string $mode
	 */
	public function __construct( string $name = '', string $mode = '' )
	{
		$this->params		=	CBAntiSpam::getGlobalParams();
		$this->name			=	$name;

		if ( ( $mode !== '' ) && ( $mode !== null ) && ( $mode !== '-1' ) ) {
			$this->mode		=	$mode;
		} else {
			$this->mode		=	$this->params->getString( 'captcha_mode', '' );
		}
	}

	/**
	 * Reloads the captcha from session by id
	 *
	 * @param string $id
	 * @return bool
	 */
	public function load( string $id = '' ): bool
	{
		if ( ! $id ) {
			$id			=	$this->getId();
		}

		if ( $id === 'preview' ) {
			$this->generateCode();

			return true;
		}

		$session		=	Application::Session()->subTree( 'captcha.' . $id );

		if ( ! $session->count() ) {
			return false;
		}

		$this->name		=	$session->getString( 'name', '' );
		$this->mode		=	$session->getString( 'mode', '' );
		$this->code		=	$session->getString( 'code', '' );
		$this->answer	=	$session->getString( 'answer', '' );

		return true;
	}

	/**
	 * Resets the captcha code
	 */
	public function reset(): void
	{
		$this->code		=	'';
		$this->answer	=	'';
		$this->error	=	'';

		$this->cache();
	}

	/**
	 * Caches the captcha into session; this is normally only done on creation or parse to preserve parameters between loads
	 * It is not advised to call this manually unless captcha parameters have changed after creation and desired result is for them to persist
	 */
	public function cache(): void
	{
		if ( ( ! $this->getId() ) || ( $this->getName() === 'preview' ) ) {
			return;
		}

		Application::Session()->set( 'captcha.' . $this->getId(), [	'name'		=>	$this->getName(),
																	'mode'		=>	$this->getMode(),
																	'code'		=>	$this->getCode(),
																	'answer'	=>	$this->getAnswer(),
																] );
	}

	/**
	 * Gets the captcha id
	 *
	 * @return string
	 */
	public function getId(): string
	{
		if ( $this->getName() === 'preview' ) {
			return 'preview';
		}

		if ( ! $this->getName() ) {
			return '';
		}

		return md5( 'captcha:' . $this->getName() );
	}

	/**
	 * Gets the captcha name
	 *
	 * @return string
	 */
	public function getName(): string
	{
		if ( ! $this->name ) {
			return '';
		}

		return strtolower( trim( preg_replace( '/[^a-zA-Z0-9.]/i', '', $this->name ) ) );
	}

	/**
	 * Gets the captcha mode
	 *
	 * @return string
	 */
	public function getMode(): string
	{
		return $this->mode;
	}

	/**
	 * Generates the captcha code
	 */
	public function generateCode(): void
	{
		if ( $this->getCode() ) {
			return;
		}

		$code							=	'';
		$answer							=	'';

		switch( $this->getMode() ) {
			case 'joomla':
			case 'recaptcha':
			case 'recaptcha_score':
			case 'recaptcha_invisible':
			case 'honeypot':
				break;
			case 'question':
				$captchaQuestions		=	"What is 2 plus 2?=4\n"
										.	"What is 1 times 6?=6\n"
										.	"What is 9 divide 3?=3\n"
										.	"Are you a Human?=Yes\n"
										.	"Are you a Bot?=No\n"
										.	"How many words is this?=5\n"
										.	"How many fingers on a hand?=5\n"
										.	"How many toes on a foot?=5\n"
										.	"What is 10 add 10?=20\n"
										.	"What is 0 multiply 100?=0\n"
										.	"What is 5 minus 1?=4\n"
										.	"What is 2 add 2?=4\n"
										.	"4th letter of Test is?=t\n"
										.	"20, 81, 3; which is smallest?=3\n"
										.	"12, 31, 9; which is greatest?=31\n"
										.	"Purple, car, dog; which is a color?=Purple\n"
										.	"Cat, plane, rock; which is an animal?=Cat\n"
										.	"If tomorrow is Monday; what day is today?=Sunday\n"
										.	"Tim, cat, dog; which is human?=Tim";

				$questions				=	$this->params->getString( 'captcha_internal_questions', $captchaQuestions );

				if ( ! $questions ) {
					$questions			=	$captchaQuestions;
				}

				$questions				=	explode( "\n", $questions );
				$codes					=	[];

				foreach ( $questions as $question ) {
					$question			=	explode( '=', $question );
					$key				=	( isset( $question[0] ) ? trim( CBTxt::T( $question[0] ) ) : null );
					$value				=	( isset( $question[1] ) ? trim( CBTxt::T( $question[1] ) ) : null );

					if ( $key && $value ) {
						$codes[$key]	=	$value;
					}
				}

				if ( $codes ) {
					$code				=	array_rand( $codes );
					$answer				=	$codes[$code];
				}
				break;
			case 'math':
				$leftMin				=	$this->params->getInt( 'captcha_internal_math_left_min', 1 );
				$leftMax				=	$this->params->getInt( 'captcha_internal_math_left_max', 10 );
				$rightMin				=	$this->params->getInt( 'captcha_internal_math_left_min', 1 );
				$rightMax				=	$this->params->getInt( 'captcha_internal_math_right_max', 10 );
				$answer					=	0;

				$code					=	CBAntiSpam::getMathEquation( $answer, $leftMin, $leftMax, $rightMin, $rightMax );
				break;
			case 'internal':
			default:
				$length					=	$this->params->getInt( 'captcha_internal_length', 6 );

				if ( ! $length ) {
					$length				=	6;
				}

				$characters				=	$this->params->getString( 'captcha_internal_characters', 'abcdefhijklmnopqrstuvwxyz' );

				if ( ! $characters ) {
					$characters			=	'abcdefhijklmnopqrstuvwxyz';
				}

				for ( $i = 0, $n = (int) $length; $i < $n; $i++ ) {
					$code				.=	cbIsoUtf_substr( $characters, random_int( 0, cbIsoUtf_strlen( $characters ) - 1 ), 1 );
				}
				break;
		}

		$this->code						=	$code;
		$this->answer					=	(string) $answer;

		$this->cache();
	}

	/**
	 * Gets the captcha code (for question based captcha this is the question)
	 *
	 * @return string
	 */
	public function getCode(): string
	{
		return $this->code;
	}

	/**
	 * Gets the captcha answer (for question based captcha)
	 *
	 * @return string
	 */
	public function getAnswer(): string
	{
		return $this->answer;
	}

	/**
	 * Gets the captcha validation code
	 *
	 * @return string
	 */
	public function getError(): string
	{
		return $this->error;
	}

	/**
	 * Returns the user supplied captcha value
	 *
	 * @return string
	 */
	public function getValue(): string
	{
		if ( $this->getName() === 'preview' ) {
			return '';
		}

		switch( $this->getMode() ) {
			case 'recaptcha':
			case 'recaptcha_score':
			case 'recaptcha_invisible':
				$value		=	Application::Input()->getString( 'g-recaptcha-response', '' );
				break;
			case 'honeypot':
				$value		=	'';
				break;
			case 'joomla':
			case 'internal':
			case 'question':
			case 'math':
			default:
				$value		=	Application::Input()->getString( $this->getId(), '' );
				break;
		}

		return $value;
	}

	/**
	 * Validates the captcha from POST
	 *
	 * @param null|string $value
	 * @return bool
	 */
	public function validateValue( ?string $value = null ): bool
	{
		global $_CB_framework;

		if ( $this->getName() === 'preview' ) {
			return false;
		}

		$reset											=	true;
		$valid											=	false;

		if ( $value !== null ) {
			$reset										=	false;
		} else {
			$value										=	$this->getValue();
		}

		$ipAddress										=	Application::Input()->getRequestIP();
		$useHoneypot									=	$this->params->getBool( 'captcha_honeypot', true );

		switch( $this->getMode() ) {
			case 'joomla':
				$jPlugin								=	$this->params->getString( 'captcha_joomla_plugin', '' );

				if ( ! $jPlugin ) {
					$jPlugin							=	$_CB_framework->getCfg( 'captcha' );
				}

				if ( $jPlugin ) {
					$jCaptcha							=	\JCaptcha::getInstance( $jPlugin, [ 'namespace' => $this->getName() ] );

					if ( $jCaptcha ) {
						$valid							=	$jCaptcha->checkAnswer( $value );
					}
				}
				break;
			case 'recaptcha':
			case 'recaptcha_score':
			case 'recaptcha_invisible':
				// v2
				$recaptchaKey							=	trim( $this->params->getString( 'captcha_recaptcha_site_key', '' ) );
				$recaptchaSecret						=	trim( $this->params->getString( 'captcha_recaptcha_secret_key', '' ) );

				// v3
				$recaptchaScoreKey						=	trim( $this->params->getString( 'captcha_recaptcha_score_site_key', '' ) );
				$recaptchaScoreSecret					=	trim( $this->params->getString( 'captcha_recaptcha_score_secret_key', '' ) );

				// invisible
				$recaptchaInvisibleKey					=	trim( $this->params->getString( 'captcha_recaptcha_invisible_site_key', '' ) );
				$recaptchaInvisibleSecret				=	trim( $this->params->getString( 'captcha_recaptcha_invisible_secret_key', '' ) );

				if ( ( ( $this->getMode() === 'recaptcha' ) && $recaptchaKey && $recaptchaSecret )
					 || ( ( $this->getMode() === 'recaptcha_score' ) && $recaptchaScoreKey && $recaptchaScoreSecret )
					 || ( ( $this->getMode() === 'recaptcha_invisible' ) && $recaptchaInvisibleKey && $recaptchaInvisibleSecret )
				) {
					try {
						$client							=	new \GuzzleHttp\Client();

						$body							=	[	'secret'	=>	( $this->getMode() === 'recaptcha' ? $recaptchaSecret : ( $this->getMode() === 'recaptcha_score' ? $recaptchaScoreSecret : $recaptchaInvisibleSecret ) ),
																'response'	=>	$value
															];

						if ( $ipAddress ) {
							$body['remoteip']			=	$ipAddress;
						}

						$result							=	$client->get( 'https://www.google.com/recaptcha/api/siteverify', [ 'query' => $body ] );

						if ( (int) $result->getStatusCode() === 200 ) {
							$response					=	json_decode( (string) $result->getBody(), true );

							if ( isset( $response['success'] ) && ( $response['success'] === true ) ) {
								if ( $this->getMode() === 'recaptcha_score' ) {
									$score				=	( isset( $response['score'] ) ? (float) $response['score'] : 0.0 );

									if ( $score >= $this->params->getFloat( 'captcha_recaptcha_score', 0.5 ) ) {
										$valid			=	true;
									}
								} else {
									$valid				=	true;
								}
							} elseif ( isset( $response['error-codes'] ) ) {
								$this->error			=	implode( ', ', $response['error-codes'] );
							}
						} else {
							$this->error				=	CBTxt::T( 'Failed to reach Google reCaptcha verify server.' );
						}
					} catch ( \Exception $e ) {
						$this->error					=	$e->getMessage();
					}
				} else {
					$valid								=	true;
				}
				break;
			case 'honeypot':
				$useHoneypot							=	true;
				$valid									=	true;
				break;
			case 'question':
			case 'math':
				$answer									=	$this->getAnswer();

				if ( $answer && ( cbutf8_strtolower( $answer ) === cbutf8_strtolower( $value ) ) ) {
					$valid								=	true;
				}
				break;
			case 'internal':
			default:
				$code									=	$this->getCode();

				if ( $code && ( cbutf8_strtolower( $code ) === cbutf8_strtolower( $value ) ) ) {
					$valid								=	true;
				}
				break;
		}

		if ( $useHoneypot ) {
			$honeypot									=	$this->params->getString( 'captcha_honeypot_name', 'full_address' );

			if ( ! $honeypot ) {
				$honeypot								=	'full_address';
			}

			if ( Application::Input()->getRaw( $honeypot ) ) {
				$valid									=	false;
			}
		}

		if ( $reset ) {
			$this->reset();

			$blocked									=	CBAntiSpam::getBlock( null, $ipAddress );
			$message									=	$this->params->get( 'captcha_autoblock_msg', 'Your captcha attempt has been blocked. Reason: [reason]' );

			if ( $blocked ) {
				if ( $message ) {
					$extras								=	[	'[duration]'	=>	ucwords( strtolower( str_replace( [ '+', '-' ], '', $blocked->getDuration() ) ) ),
																'[date]'		=>	$blocked->getDate() . ' UTC',
																'[expire]'		=>	$blocked->getExpirationDate() . ( $blocked->getDuration() ? ' UTC' : null )
															];

					$extras								=	array_merge( $extras, [ '[reason]' => CBTxt::T( 'CAPTCHA_BLOCK_REASON', ( $blocked->getReason() ?: 'Spam.' ), $extras ) ] );

					$this->error						=	CBTxt::T( 'CAPTCHA_BLOCK_MESSAGE', $message, $extras );
				}

				$valid									=	false;
			} elseif ( ! $valid ) {
				$attempt								=	CBAntiSpam::logAttempt( 'captcha' );

				if ( $attempt ) {
					$blocked							=	CBAntiSpam::blockAttempt( $attempt );
				}

				if ( $blocked && $message ) {
					$extras								=	[	'[duration]'	=>	ucwords( strtolower( str_replace( [ '+', '-' ], '', $blocked->getDuration() ) ) ),
																'[date]'		=>	$blocked->getDate() . ' UTC',
																'[expire]'		=>	$blocked->getExpirationDate() . ( $blocked->getDuration() ? ' UTC' : null )
															];

					$extras								=	array_merge( $extras, [ '[reason]' => CBTxt::T( 'CAPTCHA_BLOCK_REASON', ( $blocked->getReason() ?: 'Spam.' ), $extras ) ] );

					$this->error						=	CBTxt::T( 'CAPTCHA_BLOCK_MESSAGE', $message, $extras );
				}
			} else {
				CBAntiSpam::logIPAddress( 0, 'captcha' );
			}
		}

		return $valid;
	}

	/**
	 * Returns the captcha generated image
	 *
	 * @return string
	 */
	public function getImage(): string
	{
		global $_CB_framework, $_PLUGINS;

		CBAntiSpam::getTemplate( 'captcha' );

		static $absPath				=	null;

		if ( ! $absPath ) {
			$plugin					=	$_PLUGINS->getLoadedPlugin( 'user', 'cbantispam' );
			$absPath				=	$_PLUGINS->getPluginPath( $plugin );
		}

		switch( $this->getMode() ) {
			case 'joomla':
			case 'recaptcha':
			case 'recaptcha_score':
			case 'recaptcha_invisible':
			case 'honeypot':
				$image				=	'';
				break;
			case 'internal':
			case 'question':
			case 'math':
			default:
				if ( $this->isAjax() && ( ! $this->getCode() ) ) {
					$image			=	'<span class="cbantispamCaptchaImage" data-cbantispam-captcha="' . htmlspecialchars( $this->getId() ) . '"></span>';
				} else {
					$height			=	$this->params->getInt( 'captcha_internal_height', 60 );

					if ( ! $height ) {
						$height		=	60;
					}

					$image			=	'<img src="' . $_CB_framework->pluginClassUrl( 'cbantispam', true, [ 'action' => 'captcha', 'func' => 'image', 'id' => $this->getId() ], 'raw', 0, true ) . '" alt="' . htmlspecialchars( CBTxt::T( 'Captcha' ) ) . '"' . ( $height ? ' style="height: ' . (int) $height . 'px;"' : null ) . ' class="mw-100 cbantispamCaptchaImage cbantispamCaptchaImage' . htmlspecialchars( $this->getId() ) . '" />';
				}
				break;
		}

		return $image;
	}

	/**
	 * Returns the captcha generated audio
	 *
	 * @return string
	 */
	public function getAudio(): string
	{
		global $_CB_framework;

		if ( $this->getName() === 'preview' ) {
			return '';
		}

		CBAntiSpam::getTemplate( 'captcha' );

		switch( $this->getMode() ) {
			case 'joomla':
			case 'recaptcha':
			case 'recaptcha_score':
			case 'recaptcha_invisible':
			case 'honeypot':
				$audio		=	'';
				break;
			case 'internal':
			case 'question':
			case 'math':
			default:
				if ( $this->isAjax()  && ( ! $this->getCode() ) ) {
					$audio	=	'<span class="cbantispamCaptchaAudioFile" data-cbantispam-captcha="' . htmlspecialchars( $this->getId() ) . '"></span>';
				} else {
					$audio	=	'<audio src="' . $_CB_framework->pluginClassUrl( 'cbantispam', true, [ 'action' => 'captcha', 'func' => 'audio', 'id' => $this->getId() ], 'raw', 0, true ) . '" type="audio/mpeg" class="cbantispamCaptchaAudioFile cbantispamCaptchaAudioFile' . htmlspecialchars( $this->getId() ) . ' hidden"></audio>';
				}
				break;
		}

		return $audio;
	}

	/**
	 * @param array $data
	 * @return void
	 */
	public function setInputData( array $data ): void
	{
		$this->input	=	$data;
	}

	/**
	 * Returns the captcha form input
	 *
	 * @return string
	 */
	public function getInput(): string
	{
		CBAntiSpam::getTemplate( 'captcha' );

		if ( ! $this->isAjax() ) {
			$this->generateCode();
		} else {
			$this->cache();
		}

		$attributes						=	( $this->input['attributes'] ?? '' );
		$classes						=	( $this->input['classes '] ?? 'form-control' );
		$useHoneypot					=	$this->params->getBool( 'captcha_honeypot', true );

		switch( $this->getMode() ) {
			case 'joomla':
			case 'recaptcha':
			case 'recaptcha_score':
			case 'recaptcha_invisible':
				$input					=	'';
				break;
			case 'honeypot':
				$useHoneypot			=	true;
				$input					=	'';
				break;
			case 'internal':
			case 'question':
			case 'math':
			default:
				if ( $this->isAjax() && ( ! $this->getCode() ) ) {
					$input				=	'<span class="cbantispamCaptchaInput" data-cbantispam-captcha="' . htmlspecialchars( $this->getId() ) . '"></span>';
				} else {
					$size				=	$this->params->getInt( 'captcha_internal_size', 20 );

					if ( ! $size ) {
						$classes		.=	' w-100';
					}

					$input				=	'<input type="text" name="' . htmlspecialchars( $this->getId() ) . '" value=""' . ( $size ? ' size="' . (int) $size . '"' : null ) . ' class="' . ( $classes ? htmlspecialchars( $classes ) . ' ' : null ) . 'cbantispamCaptchaInput cbantispamCaptchaInput' . htmlspecialchars( $this->getId() ) . ' required"' . ( $attributes ? ' ' . trim( $attributes ) : null ) . ' aria-label="' . htmlspecialchars( CBTxt::T( 'Input Captcha' ) ) . '">';
				}
				break;
		}

		if ( $useHoneypot && ( ( ! $this->isAjax() ) || ( $this->isAjax() && ( ! $this->getCode() ) ) ) ) {
			$honeypot					=	$this->params->getString( 'captcha_honeypot_name', 'full_address' );

			if ( ! $honeypot ) {
				$honeypot				=	'full_address';
			}

			$input						.=	'<div class="cbantispamCaptchaInput">'
										.		'<input type="text" name="' . htmlspecialchars( $honeypot ) . '" value="" class="form-control">'
										.	'</div>';
		}

		return $input;
	}

	/**
	 * Displays the captcha
	 *
	 * @param bool $withInput
	 * @return string
	 */
	public function getCaptcha( bool $withInput = true ): string
	{
		global $_CB_framework;

		CBAntiSpam::getTemplate( 'captcha' );

		if ( ! $this->isAjax() ) {
			$this->generateCode();
		} else {
			$this->cache();
		}

		// v2
		$recaptchaKey								=	trim( $this->params->getString( 'captcha_recaptcha_site_key', '' ) );
		$recaptchaSecret							=	trim( $this->params->getString( 'captcha_recaptcha_secret_key', '' ) );

		// v3
		$recaptchaScoreKey							=	trim( $this->params->getString( 'captcha_recaptcha_score_site_key', '' ) );
		$recaptchaScoreSecret						=	trim( $this->params->getString( 'captcha_recaptcha_score_secret_key', '' ) );
		$recaptchaScoreBadge						=	$this->params->getString( 'captcha_recaptcha_score_badge', 'inline' );

		if ( $recaptchaScoreBadge === 'none' ) {
			$recaptchaScoreBadge					=	'inline';
		}

		// invisible
		$recaptchaInvisibleKey						=	trim( $this->params->getString( 'captcha_recaptcha_invisible_site_key', '' ) );
		$recaptchaInvisibleSecret					=	trim( $this->params->getString( 'captcha_recaptcha_invisible_secret_key', '' ) );
		$recaptchaInvisibleBadge					=	$this->params->getString( 'captcha_recaptcha_invisible_badge', 'inline' );

		if ( $recaptchaInvisibleBadge === 'none' ) {
			$recaptchaInvisibleBadge				=	'inline';
		}

		static $JS_loaded							=	0;

		if ( ! $JS_loaded++ ) {
			// Always load ALL configuration regardless of captcha type as another captcha on the page might be of different type and we want to load this JS only one time
			$captchaConfig							=	[	'url'		=>	$_CB_framework->pluginClassUrl( 'cbantispam', false, [ 'action' => 'captcha', 'func' => 'load' ], 'raw', 0, true ),
															'recaptcha'	=>	[
																'theme'		=>	$this->params->getString( 'captcha_recaptcha_theme', 'light' ),
																'size'		=>	$this->params->getString( 'captcha_recaptcha_size', 'normal' ),
																'badge'		=>	[
																	'v3'		=>	$recaptchaScoreBadge,
																	'invisible'	=>	$recaptchaInvisibleBadge,
																],
																'v2'		=>	$recaptchaKey,
																'v3'		=>	$recaptchaScoreKey,
																'invisible'	=>	$recaptchaInvisibleKey,
															],
														];

			$_CB_framework->document->addHeadScriptUrl( '/components/com_comprofiler/plugin/user/plug_cbantispam/js/cbantispam.js', true, "const cbantispam = " . json_encode( $captchaConfig ) . ";" );
		}

		$return										=	'';

		switch( $this->getMode() ) {
			case 'joomla':
				$jPlugin							=	$this->params->getString( 'captcha_joomla_plugin', '' );

				if ( ! $jPlugin ) {
					$jPlugin						=	$_CB_framework->getCfg( 'captcha' );
				}

				if ( $jPlugin ) {
					$jCaptcha						=	\JCaptcha::getInstance( $jPlugin, [ 'namespace' => $this->getName() ] );

					if ( $jCaptcha ) {
						$return						=	$jCaptcha->display( $this->getId(), $this->getId(), ( $this->getName() !== 'preview' ? 'class="required"' : '' ) );
					}
				}
				break;
			case 'recaptcha':
			case 'recaptcha_score':
			case 'recaptcha_invisible':
				if ( $this->getMode() === 'recaptcha_invisible' ) {
					if ( ( ! $recaptchaInvisibleKey ) || ( ! $recaptchaInvisibleSecret ) ) {
						return '';
					}
				} elseif ( $this->getMode() === 'recaptcha_score' ) {
					if ( ( ! $recaptchaScoreKey ) || ( ! $recaptchaScoreSecret ) ) {
						return '';
					}
				} elseif ( ( ! $recaptchaKey ) || ( ! $recaptchaSecret ) ) {
					return '';
				}

				static $RE_loaded					=	0;

				if ( ! $RE_loaded++ ) {
					$languages						=	[	'ar', 'af', 'am', 'hy', 'az', 'eu', 'bn', 'bg', 'ca', 'zh-HK', 'zh-CN', 'zh-TW',
															'hr', 'cs', 'da', 'nl', 'en-GB', 'en', 'et', 'fil', 'fi', 'fr', 'fr-CA', 'gl', 'ka',
															'de', 'de-AT', 'de-CH', 'el', 'gu', 'iw', 'hi', 'hu', 'is', 'id', 'it', 'ja', 'kn',
															'ko', 'lo', 'lv', 'lt', 'ms', 'ml', 'mr', 'mn', 'no', 'fa', 'pl', 'pt', 'pt-BR', 'pt-PT',
															'ro', 'ru', 'sr', 'si', 'sk', 'sl', 'es', 'es-419', 'sw', 'sv', 'ta', 'te', 'th', 'tr', 'uk', 'ur', 'vi', 'zu'
														];
					$language						=	$this->params->getString( 'captcha_recaptcha_lang', 'en' );

					if ( $language === 'auto' ) {
						$languageTag				=	Application::Cms()->getLanguageTag();

						if ( \in_array( $languageTag, $languages, true ) ) {
							$language				=	$languageTag;
						} else {
							[ $languageCode ]		=	explode( '-', Application::Cms()->getLanguageTag() );

							if ( \in_array( $languageCode, $languages, true ) ) {
								$language			=	$languageCode;
							}
						}

						if ( $language === 'auto' ) {
							$language				=	'';
						}
					}

					$scheme							=	( ( isset( $_SERVER['HTTPS'] ) && ( ! empty( $_SERVER['HTTPS'] ) ) && ( $_SERVER['HTTPS'] !== 'off' ) ) ? 'https' : 'http' );

					$_CB_framework->document->addHeadScriptUrl( $scheme . '://www.google.com/recaptcha/api.js?onload=cbAntiSpamRecaptcha&render=explicit' . ( $language ? '&hl=' . $language : null ) );
				}

				if ( $this->getMode() === 'recaptcha_invisible' ) {
					$class							=	'cbantispamRecaptchaInvisible';

					if ( $this->params->getString( 'captcha_recaptcha_invisible_badge', 'inline' ) === 'none' ) {
						$class						.=	' hidden';
					}
				} elseif ( $this->getMode() === 'recaptcha_score' ) {
					$class							=	'cbantispamRecaptchaScore';

					if ( $this->params->getString( 'captcha_recaptcha_score_badge', 'inline' ) === 'none' ) {
						$class						.=	' hidden';
					}
				} else {
					$class							=	'cbantispamRecaptcha';
				}

				$return								=	'<div class="' . htmlspecialchars( $class ). ' ' . htmlspecialchars( $class . $this->getId() ) . ( $this->getName() === 'preview' ? ' preview' : null ) . '" data-cbantispam-captcha="' . htmlspecialchars( $this->getId() ) . '" style="display: inline-block;"></div>'
													.	( $withInput ? $this->getInput() : '' );
				break;
			case 'honeypot':
				$return								=	( $withInput ? $this->getInput() : '' );
				break;
			case 'internal':
			case 'question':
			case 'math':
			default:
				if ( $this->isAjax() && ( ! $this->getCode() ) ) {
					$return							=	'<span class="cbantispamCaptcha" data-cbantispam-captcha="' . htmlspecialchars( $this->getId() ) . '"><div class="cbantispamCaptchaLoading spinner-border spinner-border-sm text-secondary"></div></span>';
				} else {
					if ( $this->getName() === 'preview' ) {
						$audio						=	false;
						$refresh					=	false;
					} else {
						$audio						=	( \in_array( $this->getMode(), [ 'question', 'math' ], true ) ? false : $this->params->getBool( 'captcha_internal_audio', true ) );
						$refresh					=	$this->params->getBool( 'captcha_internal_refresh', true );
					}

					$return							=	'<div class="d-inline-flex flex-column mw-100 cbantispamCaptcha cbantispamCaptcha' . htmlspecialchars( $this->getId() ) . '" data-cbantispam-captcha="' . htmlspecialchars( $this->getId() ) . '" style="gap: 0.25rem">'
													.		'<div class="d-flex cbantispamCaptchaContainer" style="gap: 0.5rem">';

					if ( $audio || $refresh ) {
						$return						.=			'<div class="cbantispamCaptchaImageContaner">' . $this->getImage() . '</div>'
													.			'<div class="d-flex flex-column text-center cbantispamCaptchaButtons" style="gap: 0.25rem; min-width: 24px;">';

						if ( $refresh ) {
							$return					.=				'<div class="cbantispamCaptchaRefreshButton">'
													.					'<a href="javascript:void(0);" class="text-large cbantispamCaptchaRefresh cbantispamCaptchaRefresh' . htmlspecialchars( $this->getId() ) . ' fa fa-refresh" style="cursor: pointer;" title="' . htmlspecialchars( CBTxt::T( 'Refresh Captcha' ) ) . '"></a>'
													.				'</div>';
						}

						if ( $audio ) {
							$return					.=				'<div class="cbantispamCaptchaAudioButton">'
													.					'<a href="javascript:void(0);" class="text-large cbantispamCaptchaAudio cbantispamCaptchaAudio' . htmlspecialchars( $this->getId() ) . ' fa fa-volume-up" style="cursor: pointer;" title="' . htmlspecialchars( CBTxt::T( 'Listen to Captcha' ) ) . '"></a>'
													.					$this->getAudio()
													.				'</div>';
						}

						$return						.=			'</div>';
					} else {
						$return						.=			$this->getImage();
					}

					$return							.=		'</div>'
													.		( $withInput ? $this->getInput() : '' )
													.	'</div>';
				}
				break;
		}

		return $return;
	}

	/**
	 * @return bool
	 */
	private function isAjax(): bool
	{
		return ( $this->params->getBool( 'captcha_internal_ajax', false ) && ( $this->getName() !== 'preview' ) && \in_array( $this->getMode(), [ 'internal', 'question', 'math' ], true ) );
	}
}