<?php
/**
* Community Builder (TM)
* @version $Id: $
* @package CommunityBuilder
* @copyright (C) 2004-2022 www.joomlapolis.com / Lightning MultiCom SA - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

namespace CB\Plugin\AntiSpam\Table;

use CB\Database\Table\UserTable;
use CBLib\Application\Application;
use CBLib\Database\Table\Table;
use CBLib\Language\CBTxt;
use CB\Plugin\AntiSpam\CBAntiSpam;
use CBuser;

\defined('CBLIB') or die();

class BlockTable extends Table
{
	/** @var null|int  */
	public $id;
	/** @var null|string  */
	public $type;
	/** @var null|string|int  */
	public $value;
	/** @var null|string  */
	public $duration;
	/** @var null|string  */
	public $reason;
	/** @var null|string  */
	public $date;

	/** @var null|string  */
	protected $_custom_duration;

	/**
	 * Table name in database
	 *
	 * @var string
	 */
	protected $_tbl					=	'#__comprofiler_plugin_antispam_block';

	/**
	 * Primary key(s) of table
	 *
	 * @var string
	 */
	protected $_tbl_key				=	'id';

	/**
	 * Copy the named array or object content into this object as vars
	 * only existing vars of object are filled.
	 * When undefined in array, object variables are kept.
	 *
	 * WARNING: DOES addslashes / escape BY DEFAULT
	 *
	 * Can be overridden or overloaded.
	 *
	 * @param  array|object  $array         The input array or object
	 * @param  string        $ignore        Fields to ignore
	 * @param  string        $prefix        Prefix for the array keys
	 * @return bool                         TRUE: ok, FALSE: error on array binding
	 */
	public function bind( $array, $ignore = '', $prefix = null ): bool
	{
		$bind								=	parent::bind( $array, $ignore, $prefix );

		// Bind the custom duration to duration if it exists
		if ( $bind ) {
			if ( \is_array( $array ) && isset( $array['_custom_duration'] ) ) {
				$this->_custom_duration		=	$array['_custom_duration'];
			} elseif ( isset( $array->_custom_duration ) ) {
				$this->_custom_duration		=	$array->_custom_duration;
			}

			if ( $this->_custom_duration !== '' ) {
				$this->duration				=	$this->_custom_duration;
				$this->_custom_duration		=	null;
			}
		}

		return $bind;
	}

	/**
	 * @return bool
	 */
	public function check(): bool
	{
		if ( $this->getType() === '' ) {
			$this->setError( CBTxt::T( 'Type not specified!' ) );

			return false;
		}

		if ( $this->getValue() === '' ) {
			$this->setError( CBTxt::T( 'Value not specified!' ) );

			return false;
		}

		if ( ! \in_array( $this->getType(), [ 'account', 'user', 'ban', 'reset', 'ip', 'ip_partial', 'ip_range', 'ip_list', 'email', 'email_partial', 'email_list', 'domain', 'domain_partial', 'domain_list' ], true ) ) {
			$this->setError( CBTxt::T( 'Type not valid!' ) );

			return false;
		}

		if ( ( $this->getType() === 'email' ) && ( strpos( $this->getValue(), '%' ) === false ) && ( ! cbIsValidEmail( $this->getValue() ) ) ) {
			$this->setError( CBTxt::T( 'Email Address not valid!' ) );

			return false;
		}

		if ( \in_array( $this->getType(), [ 'account', 'user', 'reset' ], true ) ) {
			if ( ! $this->getUser()->getInt( 'id', 0 ) ) {
				$this->setError( CBTxt::T( 'User does not exist!' ) );

				return false;
			}

			if ( ( $this->getType() === 'account' ) && $this->getUser()->getInt( 'block', 0 ) ) {
				$this->setError( CBTxt::T( 'User already blocked!' ) );

				return false;
			}

			if ( ( $this->getType() === 'user' ) && $this->getUser()->getInt( 'banned', 0 ) ) {
				$this->setError( CBTxt::T( 'User already banned!' ) );

				return false;
			}

			if ( ( $this->getType() === 'reset' ) && $this->getUser()->getInt( 'requireReset', 0 ) ) {
				$this->setError( CBTxt::T( 'User already reset!' ) );

				return false;
			}
		}

		return true;
	}

	/**
	 * @param bool $updateNulls
	 * @return bool
	 */
	public function store( $updateNulls = false ): bool
	{
		global $_PLUGINS;

		if ( $this->getType() === 'account' ) {
			return $this->blockUser();
		}

		if ( $this->getType() === 'ban' ) {
			return $this->banUser();
		}

		if ( $this->getType() === 'reset' ) {
			return $this->resetUser();
		}

		$new	=	( ! $this->getId() );
		$old	=	new self();

		$this->setDate( $this->getString( 'date', Application::Database()->getUtcDateTime() ) );

		if ( ! $new ) {
			$old->load( $this->getId() );

			$_PLUGINS->trigger( 'antispam_onBeforeUpdateBlock', [ &$this, $old ] );
		} else {
			$_PLUGINS->trigger( 'antispam_onBeforeCreateBlock', [ &$this ] );
		}

		if ( ! parent::store( $updateNulls ) ) {
			return false;
		}

		if ( ! $new ) {
			$_PLUGINS->trigger( 'antispam_onAfterUpdateBlock', [ $this, $old ] );
		} else {
			$_PLUGINS->trigger( 'antispam_onAfterCreateBlock', [ $this ] );
		}

		return true;
	}

	/**
	 * @param null|int $id
	 * @return bool
	 */
	public function delete( $id = null ): bool
	{
		global $_PLUGINS;

		$_PLUGINS->trigger( 'antispam_onBeforeDeleteBlock', [ &$this ] );

		if ( ! parent::delete( $id ) ) {
			return false;
		}

		$_PLUGINS->trigger( 'antispam_onAfterDeleteBlock', [ $this ] );

		return true;
	}

	/**
	 * @return int
	 */
	public function getId(): int
	{
		return $this->getInt( 'id', 0 );
	}

	/**
	 * @param null|int $id
	 */
	public function setId( ?int $id ): void
	{
		$this->id	=	$id;
	}

	/**
	 * @return string
	 */
	public function getType(): string
	{
		return $this->getString( 'type', '' );
	}

	/**
	 * @param string $type
	 */
	public function setType( string $type ): void
	{
		$this->type	=	$type;
	}

	/**
	 * @return string
	 */
	public function getValue(): string
	{
		return $this->getString( 'value', '' );
	}

	/**
	 * @return int
	 */
	public function getUserId(): int
	{
		if ( ! \in_array( $this->getType(), [ 'account', 'user', 'reset' ], true ) ) {
			return 0;
		}

		return $this->getInt( 'value', 0 );
	}

	/**
	 * @return string
	 */
	public function getIpAddress(): string
	{
		if ( $this->getType() !== 'ip' ) {
			return '';
		}

		return $this->getString( 'value', '' );
	}

	/**
	 * @return UserTable
	 */
	public function getUser(): UserTable
	{
		return CBuser::getUserDataInstance( $this->getUserId() );
	}

	/**
	 * @param string|int $value
	 */
	public function setValue( $value ): void
	{
		$this->value	=	$value;
	}

	/**
	 * @return string
	 */
	public function getDuration(): string
	{
		return $this->getString( 'duration', '' );
	}

	/**
	 * @param null|string $duration
	 */
	public function setDuration( ?string $duration ): void
	{
		$this->duration	=	$duration;
	}

	/**
	 * @return string
	 */
	public function getReason(): string
	{
		return $this->getString( 'reason', '' );
	}

	/**
	 * @param null|string $reason
	 */
	public function setReason( ?string $reason ): void
	{
		$this->reason	=	$reason;
	}

	/**
	 * @return string
	 */
	public function getDate(): string
	{
		return $this->getString( 'date', '' );
	}

	/**
	 * @param string $date
	 */
	public function setDate( string $date ): void
	{
		$this->date	=	$date;
	}

	/**
	 * @return string[]
	 */
	public function getList(): array
	{
		if ( ( ! \in_array( $this->getType(), [ 'ip_list', 'email_list', 'domain_list' ], true ) ) || ( ! $this->getValue() ) ) {
			return [];
		}

		return CBAntiSpam::getList( $this->getValue() );
	}

	/**
	 * @return string
	 */
	public function getExpirationDate(): string
	{
		static $cache	=	[];

		$id				=	$this->getId();

		if ( \array_key_exists( $id, $cache ) ) {
			return $cache[$id];
		}

		if ( $this->getDuration() ) {
			$expiry		=	Application::Date( $this->getDate(), 'UTC' )->modify( strtoupper( $this->getDuration() ) )->format( 'Y-m-d H:i:s' );
		} else {
			$expiry		=	'';
		}

		$cache[$id]		=	$expiry;

		return $expiry;
	}

	/**
	 * @return bool
	 */
	public function isExpired(): bool
	{
		static $cache	=	[];

		$id				=	$this->getId();

		if ( \array_key_exists( $id, $cache ) ) {
			return $cache[$id];
		}

		if ( $this->getDuration() ) {
			$expired	=	( Application::Date( 'now', 'UTC' )->getTimestamp() >= Application::Date( $this->getDate(), 'UTC' )->modify( strtoupper( $this->getDuration() ) )->getTimestamp() );
		} else {
			$expired	=	false;
		}

		$cache[$id]		=	$expired;

		return $expired;
	}

	/**
	 * @return bool
	 */
	public function blockUser(): bool
	{
		$user				=	$this->getUser();

		if ( ! $user->getInt( 'id', 0 ) ) {
			$this->setError( CBTxt::T( 'User does not exist!' ) );

			return false;
		}

		if ( $user->getInt( 'block', 0 ) ) {
			$this->setError( CBTxt::T( 'User already blocked!' ) );

			return false;
		}

		$user->set( 'block', 1 );

		if ( ! $user->storeBlock() ) {
			$this->setError( $user->getError() );

			return false;
		}

		return true;
	}

	/**
	 * @param string $reason
	 * @return bool
	 */
	public function banUser( string $reason = '' ): bool
	{
		if ( ! Application::Config()->getBool( 'allowUserBanning', true ) ) {
			$this->setError( CBTxt::T( 'User bans are not allowed!' ) );

			return false;
		}

		$user				=	$this->getUser();

		if ( ! $user->getInt( 'id', 0 ) ) {
			$this->setError( CBTxt::T( 'User does not exist!' ) );

			return false;
		}

		if ( $user->getInt( 'banned', 0 ) ) {
			$this->setError( CBTxt::T( 'User already banned!' ) );

			return false;
		}

		if ( ! $user->banUser( 1, Application::MyUser()->getUserId(), $reason ) ) {
			$this->setError( $user->getError() );

			return false;
		}

		return true;
	}

	/**
	 * @return bool
	 */
	public function resetUser(): bool
	{
		global $_CB_database;

		$user	=	$this->getUser();

		if ( ! $user->getInt( 'id', 0 ) ) {
			$this->setError( CBTxt::T( 'User does not exist!' ) );

			return false;
		}

		if ( $user->getInt( 'requireReset', 0 ) ) {
			$this->setError( CBTxt::T( 'User already reset!' ) );

			return false;
		}

		if ( ! $user->storeDatabaseValue( 'requireReset', 1 ) ) {
			$this->setError( $user->getError() );

			return false;
		}

		// Also delete any login sessions for this user to force them logged out
		$query	=	'DELETE'
				.	"\n FROM " . $_CB_database->NameQuote( '#__session' )
				.	"\n WHERE " . $_CB_database->NameQuote( 'userid' ) . " = " . $user->getInt( 'id', 0 );
		$_CB_database->setQuery( $query );
		$_CB_database->query();

		return true;
	}

	/**
	 * @return string
	 */
	public function renderType(): string
	{
		return CBAntiSpam::getType( $this->getType() );
	}

	/**
	 * @return string
	 */
	public function renderBlockReason(): string
	{
		$extras		=	[	'[duration]'	=>	ucwords( strtolower( str_replace( [ '+', '-' ], '', $this->getDuration() ) ) ),
							'[date]'		=>	$this->getDate() . ' UTC',
							'[expire]'		=>	$this->getExpirationDate() . ( $this->getDuration() ? ' UTC' : null ),
						];

		return CBTxt::T( 'ANTISPAM_BLOCK_REASON', $this->getReason(), $extras );
	}
}