<?php
/**
* Community Builder (TM)
* @version $Id: $
* @package CommunityBuilder
* @copyright (C) 2004-2023 www.joomlapolis.com / Lightning MultiCom SA - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

namespace CB\Plugin\ReconfirmEmail\Table;

use CB\Plugin\ReconfirmEmail\Helper;
use CBLib\Application\Application;
use CBLib\Database\Table\Table;
use CBLib\Language\CBTxt;

\defined( 'CBLIB' ) or die();

class ReconfirmEmailTable extends Table
{
	/** @var null|int  */
	public $id;
	/** @var null|int  */
	public $user_id;
	/** @var null|string  */
	public $from;
	/** @var null|string  */
	public $to;
	/** @var null|string  */
	public $code;
	/** @var null|string  */
	public $date;
	/** @var null|string  */
	public $status;

	/**
	 * Table name in database
	 *
	 * @var string
	 */
	protected $_tbl			=	'#__comprofiler_plugin_emails';

	/**
	 * Primary key(s) of table
	 *
	 * @var string
	 */
	protected $_tbl_key		=	'id';

	/**
	 * @return bool
	 */
	public function check(): bool
	{
		if ( $this->getString( 'from', '' ) === '' ) {
			$this->setError( CBTxt::T( 'Old email address not specified!' ) );

			return false;
		}

		if ( $this->getString( 'to', '' ) === '' ) {
			$this->setError( CBTxt::T( 'New email address not specified!' ) );

			return false;
		}

		if ( ! $this->getInt( 'user_id', 0 ) ) {
			$this->setError( CBTxt::T( 'User not specified!' ) );

			return false;
		}

		return true;
	}

	/**
	 * @param bool $updateNulls
	 * @return bool
	 */
	public function store( $updateNulls = false ): bool
	{
		global $_CB_framework, $_PLUGINS;

		$new	=	( ! $this->getInt( 'id', 0 ) );
		$old	=	new self();

		$this->set( 'date', $this->getString( 'date', Application::Database()->getUtcDateTime() ) );

		if ( PHP_VERSION_ID >= 70000 ) {
			$this->set( 'code', $this->getString( 'code', bin2hex( random_bytes( 32 ) ) ) );
		} else {
			$this->set( 'code', $this->getString( 'code', hash( 'sha256', $this->getInt( 'id', 0 ) . $this->getString( 'from', '' ) . $this->getString( 'to', '' ) . $_CB_framework->getCfg( 'secret' ) ) ) );
		}

		$this->set( 'status', $this->getString( 'status', 'P' ) );

		if ( ! $new ) {
			$old->load( $this->getInt( 'id', 0 ) );

			$_PLUGINS->trigger( 'reconfirm_onBeforeUpdateReconfirmEmail', [ &$this, $old ] );
		} else {
			$_PLUGINS->trigger( 'reconfirm_onBeforeCreateReconfirmEmail', [ &$this ] );
		}

		if ( ! parent::store( $updateNulls ) ) {
			return false;
		}

		if ( ! $new ) {
			$_PLUGINS->trigger( 'reconfirm_onAfterUpdateReconfirmEmail', [ $this, $old ] );
		} else {
			$_PLUGINS->trigger( 'reconfirm_onAfterCreateReconfirmEmail', [ $this ] );
		}

		return true;
	}

	/**
	 * @param null|int $id
	 * @return bool
	 */
	public function delete( $id = null ): bool
	{
		global $_PLUGINS;

		$_PLUGINS->trigger( 'reconfirm_onBeforeDeleteReconfirmEmail', [ &$this ] );

		if ( ! parent::delete( $id ) ) {
			return false;
		}

		$_PLUGINS->trigger( 'reconfirm_onAfterDeleteReconfirmEmail', [ $this ] );

		return true;
	}

	/**
	 * Confirms the email change
	 *
	 * @return bool
	 */
	public function confirm(): bool
	{
		if ( $this->getString( 'status', '' ) === 'A' ) {
			return true;
		}

		$user	=	\CBuser::getUserDataInstance( $this->getInt( 'user_id', 0 ) );

		if ( ! $user->storeDatabaseValue( 'email', $this->getString( 'to', '' ) ) ) {
			$this->setError( $user->getError() );

			return false;
		}

		$this->set( 'status', 'A' );

		if ( $this->getError() || ( ! $this->check() ) ) {
			return false;
		}

		if ( $this->getError() || ( ! $this->store() ) ) {
			return false;
		}

		return true;
	}

	/**
	 * Cancels the email change
	 *
	 * @return bool
	 */
	public function cancel(): bool
	{
		if ( $this->getString( 'status', '' ) === 'X' ) {
			return true;
		}

		$this->set( 'status', 'X' );

		if ( $this->getError() || ( ! $this->check() ) ) {
			return false;
		}

		if ( $this->getError() || ( ! $this->store() ) ) {
			return false;
		}

		return true;
	}

	/**
	 * Checks if email change has expired and sets expired status if it has
	 *
	 * @return bool
	 */
	public function expired(): bool
	{
		static $cache		=	[];

		if ( $this->getString( 'status', '' ) === 'E' ) {
			return true;
		}

		$timeframe			=	Helper::getGlobalParams()->getString( 'reconfirm_expired_timeframe', '+1 WEEK' );

		if ( ! $timeframe ) {
			return false;
		}

		$id					=	$this->getInt( 'id', 0 );

		if ( ! isset( $cache[$id] ) ) {
			$expired		=	( Application::Date( 'now', 'UTC' )->getTimestamp() >= Application::Date( $this->getString( 'date', '' ), 'UTC' )->modify( strtoupper( $timeframe ) )->getTimestamp() );

			if ( $expired ) {
				$this->set( 'status', 'E' );

				$this->store();
			}

			$cache[$id]		=	$expired;
		}

		return $cache[$id];
	}
}