<?php
/**
* Community Builder (TM)
* @version $Id: $
* @package CommunityBuilder
* @copyright (C) 2004-2022 www.joomlapolis.com / Lightning MultiCom SA - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

namespace CB\Plugin\AntiSpam\Table;

use CB\Database\Table\UserTable;
use CB\Plugin\AntiSpam\CBAntiSpam;
use CBLib\Application\Application;
use CBLib\Database\Table\Table;
use CBLib\Language\CBTxt;
use CBuser;

\defined('CBLIB') or die();

class WhitelistTable extends Table
{
	/** @var null|int  */
	public $id;
	/** @var null|string  */
	public $type;
	/** @var null|string|int  */
	public $value;
	/** @var null|string  */
	public $reason;
	/** @var null|string  */
	public $date;

	/**
	 * Table name in database
	 *
	 * @var string
	 */
	protected $_tbl			=	'#__comprofiler_plugin_antispam_whitelist';

	/**
	 * Primary key(s) of table
	 *
	 * @var string
	 */
	protected $_tbl_key		=	'id';

	/**
	 * @return bool
	 */
	public function check(): bool
	{
		if ( $this->getType() === '' ) {
			$this->setError( CBTxt::T( 'Type not specified!' ) );

			return false;
		}

		if ( $this->getValue() === '' ) {
			$this->setError( CBTxt::T( 'Value not specified!' ) );

			return false;
		}

		if ( ! \in_array( $this->getType(), [ 'account', 'user', 'ban', 'reset', 'ip', 'ip_partial', 'ip_range', 'ip_list', 'email', 'email_partial', 'email_list', 'domain', 'domain_partial', 'domain_list' ], true ) ) {
			$this->setError( CBTxt::T( 'Type not valid!' ) );

			return false;
		}

		if ( ( $this->getType() === 'email' ) && ( strpos( $this->getValue(), '%' ) === false ) && ( ! cbIsValidEmail( $this->getValue() ) ) ) {
			$this->setError( CBTxt::T( 'Email Address not valid!' ) );

			return false;
		}

		if ( \in_array( $this->getType(), [ 'account', 'user', 'reset' ], true ) ) {
			if ( ! $this->getUser()->getInt( 'id', 0 ) ) {
				$this->setError( CBTxt::T( 'User does not exist!' ) );

				return false;
			}

			if ( ( $this->getType() === 'account' ) && $this->getUser()->getInt( 'block', 0 ) ) {
				$this->setError( CBTxt::T( 'User already blocked!' ) );

				return false;
			}

			if ( ( $this->getType() === 'user' ) && $this->getUser()->getInt( 'banned', 0 ) ) {
				$this->setError( CBTxt::T( 'User already banned!' ) );

				return false;
			}

			if ( ( $this->getType() === 'reset' ) && $this->getUser()->getInt( 'requireReset', 0 ) ) {
				$this->setError( CBTxt::T( 'User already reset!' ) );

				return false;
			}
		}

		return true;
	}

	/**
	 * @param bool $updateNulls
	 * @return bool
	 */
	public function store( $updateNulls = false ): bool
	{
		global $_PLUGINS;

		if ( \in_array( $this->getType(), [ 'account', 'ban', 'reset' ], true ) ) {
			$this->setType( 'user' );
		}

		$new	=	( ! $this->getId() );
		$old	=	new self();

		$this->setDate( $this->getString( 'date', Application::Database()->getUtcDateTime() ) );

		if ( ! $new ) {
			$old->load( $this->getId() );

			$_PLUGINS->trigger( 'antispam_onBeforeUpdateWhitelist', [ &$this, $old ] );
		} else {
			$_PLUGINS->trigger( 'antispam_onBeforeCreateWhitelist', [ &$this ] );
		}

		if ( ! parent::store( $updateNulls ) ) {
			return false;
		}

		if ( ! $new ) {
			$_PLUGINS->trigger( 'antispam_onAfterUpdateWhitelist', [ $this, $old ] );
		} else {
			$_PLUGINS->trigger( 'antispam_onAfterCreateWhitelist', [ $this ] );
		}

		return true;
	}

	/**
	 * @param null|int $id
	 * @return bool
	 */
	public function delete( $id = null ): bool
	{
		global $_PLUGINS;

		$_PLUGINS->trigger( 'antispam_onBeforeDeleteWhitelist', [ &$this ] );

		if ( ! parent::delete( $id ) ) {
			return false;
		}

		$_PLUGINS->trigger( 'antispam_onAfterDeleteWhitelist', [ $this ] );

		return true;
	}

	/**
	 * @return int
	 */
	public function getId(): int
	{
		return $this->getInt( 'id', 0 );
	}

	/**
	 * @param null|int $id
	 */
	public function setId( ?int $id ): void
	{
		$this->id	=	$id;
	}

	/**
	 * @return string
	 */
	public function getType(): string
	{
		return $this->getString( 'type', '' );
	}

	/**
	 * @param string $type
	 */
	public function setType( string $type ): void
	{
		$this->type	=	$type;
	}

	/**
	 * @return string
	 */
	public function getValue(): string
	{
		return $this->getString( 'value', '' );
	}

	/**
	 * @return int
	 */
	public function getUserId(): int
	{
		if ( ! \in_array( $this->getType(), [ 'account', 'user', 'reset' ], true ) ) {
			return 0;
		}

		return $this->getInt( 'value', 0 );
	}

	/**
	 * @return string
	 */
	public function getIpAddress(): string
	{
		if ( $this->getType() !== 'ip' ) {
			return '';
		}

		return $this->getString( 'value', '' );
	}

	/**
	 * @return UserTable
	 */
	public function getUser(): UserTable
	{
		return CBuser::getUserDataInstance( $this->getUserId() );
	}

	/**
	 * @param string|int $value
	 */
	public function setValue( $value ): void
	{
		$this->value	=	$value;
	}

	/**
	 * @return string
	 */
	public function getReason(): string
	{
		return $this->getString( 'reason', '' );
	}

	/**
	 * @param null|string $reason
	 */
	public function setReason( ?string $reason ): void
	{
		$this->reason	=	$reason;
	}

	/**
	 * @return string
	 */
	public function getDate(): string
	{
		return $this->getString( 'date', '' );
	}

	/**
	 * @param string $date
	 */
	public function setDate( string $date ): void
	{
		$this->date	=	$date;
	}

	/**
	 * @return string[]
	 */
	public function getList(): array
	{
		if ( ( ! \in_array( $this->getType(), [ 'ip_list', 'email_list', 'domain_list' ], true ) ) || ( ! $this->getValue() ) ) {
			return [];
		}

		return CBAntiSpam::getList( $this->getValue() );
	}

	/**
	 * @return string
	 */
	public function renderType(): string
	{
		return CBAntiSpam::getType( $this->getType() );
	}
}