<?php
/**
* Community Builder (TM)
* @version $Id: $
* @package CommunityBuilder
* @copyright (C) 2004-2022 www.joomlapolis.com / Lightning MultiCom SA - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

namespace CB\Plugin\AntiSpam\Trigger;

use CB\Plugin\AntiSpam\CBAntiSpam;
use CB\Plugin\AntiSpam\Table\BlockTable;
use CBLib\Language\CBTxt;
use CB\Plugin\AntiSpam\Captcha;
use CB\Database\Table\UserTable;

\defined('CBLIB') or die();

class ForgotLoginTrigger extends \cbPluginHandler
{

	/**
	 * Blocks a forgot login attempt
	 *
	 * @param string|BlockTable $block
	 */
	private function blockForgotLogin( $block ): void
	{
		global $_PLUGINS;

		if ( \is_object( $block ) ) {
			if ( $block instanceof BlockTable ) {
				if ( $block->isExpired() ) {
					return;
				}

				$reason		=	$block->getReason();
				$duration	=	$block->getDuration();
				$date		=	$block->getDate();
				$expire		=	$block->getExpirationDate();
			} else {
				return;
			}
		} else {
			$reason			=	$block;
			$duration		=	null;
			$date			=	null;
			$expire			=	null;
		}

		if ( ! $reason ) {
			$reason			=	'Spam.';
		}

		$extras				=	[	'[duration]'	=>	ucwords( strtolower( str_replace( [ '+', '-' ], '', $duration ) ) ),
									'[date]'		=>	$date . ' UTC',
									'[expire]'		=>	$expire . ( $duration ? ' UTC' : null ),
								];

		$extras				=	array_merge( $extras, [ '[reason]' => CBTxt::T( 'FORGOT_BLOCK_REDIRECT_REASON', $reason, $extras ) ] );

		$redirect			=	$this->params->getString( 'forgot_block_redirect', '' );
		$redirectMsg		=	CBTxt::T( 'FORGOT_BLOCK_REDIRECT_MSG', $this->params->getString( 'forgot_block_redirect_msg', 'Your forgot login attempt has been blocked. Reason: [reason]' ), $extras );
		$redirectType		=	$this->params->getString( 'forgot_block_redirect_type', 'error' );

		if ( ! $redirect ) {
			$redirect		=	'index.php';
		}

		cbRedirect( $redirect, $redirectMsg, $redirectType );

		$_PLUGINS->_setErrorMSG( $redirectMsg );
		$_PLUGINS->raiseError();
	}

	/**
	 * Displays legacy forgot login captcha
	 *
	 * @return null|array
	 */
	public function displayCaptcha(): ?array
	{
		if ( ! $this->params->getBool( 'captcha_legacy_forgot', false ) ) {
			return null;
		}

		$captcha	=	new Captcha( 'login.forgot', $this->params->getString( 'captcha_legacy_forgot_mode', '' ) );

		return [ null, $captcha->getCaptcha() ];
	}

	/**
	 * Handles forgot login blocking, attempts log, and captcha validation
	 */
	public function beforeForgotLogin(): void
	{
		global $_PLUGINS;

		$blocked		=	CBAntiSpam::getBlock();

		if ( $blocked ) {
			$this->blockForgotLogin( $blocked );
			return;
		}

		$attempt		=	CBAntiSpam::logAttempt( 'forgot' );

		if ( $attempt ) {
			$blocked	=	CBAntiSpam::blockAttempt( $attempt );
		}

		if ( $blocked ) {
			$this->blockForgotLogin( $blocked );
			return;
		}

		if ( $this->params->getBool( 'captcha_legacy_forgot', false ) ) {
			$captcha	=	new Captcha( 'login.forgot', $this->params->getString( 'captcha_legacy_forgot_mode', '' ) );

			if ( ( ! $captcha->load() ) || ! ( $captcha->validateValue() ) ) {
				$_PLUGINS->_setErrorMSG( ( $captcha->getError() ?: CBTxt::T( 'Invalid Captcha Code' ) ) );
				$_PLUGINS->raiseError();
			}
		}
	}

	/**
	 * Updates users ip address log and clears attempts
	 *
	 * @param UserTable $user
	 * @param bool      $res
	 */
	public function afterForgotLoginUsername( $user, &$res ): void
	{
		if ( ! $res ) {
			return;
		}

		CBAntiSpam::logIPAddress( $user, 'forgot' );
	}

	/**
	 * Updates users ip address log and clears attempts
	 *
	 * @param UserTable $user
	 * @param string    $newpass
	 * @param bool      $res
	 */
	public function afterForgotLoginPassword( $user, $newpass, &$res ): void
	{
		if ( ! $res ) {
			return;
		}

		CBAntiSpam::logIPAddress( $user, 'forgot' );
	}
}