<?php
/**
* Community Builder (TM)
* @version $Id: $
* @package CommunityBuilder
* @copyright (C) 2004-2022 www.joomlapolis.com / Lightning MultiCom SA - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

namespace CB\Plugin\AntiSpam\Trigger;

use CBLib\Application\Application;
use CB\Database\Table\UserTable;
use CB\Plugin\AntiSpam\CBAntiSpam;
use CB\Plugin\AntiSpam\Table\BlockTable;
use CBLib\Language\CBTxt;
use CB\Plugin\AntiSpam\Captcha;

\defined('CBLIB') or die();

class LoginTrigger extends \cbPluginHandler
{

	/**
	 * Blocks a login attempt
	 *
	 * @param string|BlockTable $block
	 */
	private function blockLogin( $block ): void
	{
		global $_PLUGINS;

		if ( \is_object( $block ) ) {
			if ( $block instanceof BlockTable ) {
				if ( $block->isExpired() ) {
					return;
				}

				$reason		=	$block->getReason();
				$duration	=	$block->getDuration();
				$date		=	$block->getDate();
				$expire		=	$block->getExpirationDate();
			} else {
				return;
			}
		} else {
			$reason			=	$block;
			$duration		=	null;
			$date			=	null;
			$expire			=	null;
		}

		if ( ! $reason ) {
			$reason			=	'Spam.';
		}

		$extras				=	[	'[duration]'	=>	ucwords( strtolower( str_replace( [ '+', '-' ], '', $duration ) ) ),
									'[date]'		=>	$date . ' UTC',
									'[expire]'		=>	$expire . ( $duration ? ' UTC' : null ),
								];

		$extras				=	array_merge( $extras, [ '[reason]' => CBTxt::T( 'LOGIN_BLOCK_REDIRECT_REASON', $reason, $extras ) ] );

		$redirect			=	$this->params->getString( 'login_block_redirect', '' );
		$redirectMsg		=	CBTxt::T( 'LOGIN_BLOCK_REDIRECT_MSG', $this->params->getString( 'login_block_redirect_msg', 'Your login attempt has been blocked. Reason: [reason]' ), $extras );
		$redirectType		=	$this->params->getString( 'login_block_redirect_type', 'error' );

		if ( ! $redirect ) {
			$redirect		=	'index.php';
		}

		cbRedirect( $redirect, $redirectMsg, $redirectType );

		$_PLUGINS->_setErrorMSG( $redirectMsg );
		$_PLUGINS->raiseError();
	}

	/**
	 * Displays legacy login captcha
	 *
	 * @return null|string
	 */
	public function displayCaptcha(): ?string
	{
		if ( ! $this->params->getBool( 'captcha_legacy_login', false ) ) {
			return null;
		}

		$captcha	=	new Captcha( 'login', $this->params->getString( 'captcha_legacy_login_mode', '' ) );

		return	'<div class="cbLegacyLoginCaptcha cb_template cb_template_' . selectTemplate( 'dir' ) . '">'
			.		$captcha->getCaptcha()
			.	'</div>';
	}

	/**
	 * Handles login blocking (only ip address), attempts log, and captcha validation
	 *
	 * @param string      $username
	 * @param string|bool $password
	 * @param string      $secretKey
	 */
	public function beforeLogin( &$username, &$password, &$secretKey ): void
	{
		global $_PLUGINS;

		$blocked		=	CBAntiSpam::getBlock();

		if ( $blocked ) {
			$this->blockLogin( $blocked );
			return;
		}

		$attempt		=	CBAntiSpam::logAttempt( 'login' );

		if ( $attempt ) {
			$blocked	=	CBAntiSpam::blockAttempt( $attempt );
		}

		if ( $blocked ) {
			$this->blockLogin( $blocked );
			return;
		}

		if ( $this->params->getBool( 'captcha_legacy_login', false ) && $username && $password ) {
			$captcha	=	new Captcha( 'login', $this->params->getString( 'captcha_legacy_login_mode', '' ) );

			if ( ( ! $captcha->load() ) || ( ! $captcha->validateValue() ) ) {
				$_PLUGINS->_setErrorMSG( ( $captcha->getError() ?: CBTxt::T( 'Invalid Captcha Code' ) ) );
				$_PLUGINS->raiseError();
			}
		}
	}

	/**
	 * Handles login blocking (user and ip address), duplicate login check, and login share check
	 *
	 * @param UserTable   $user
	 * @param int         $foundUser
	 * @param null|string $returnPluginsOverrides
	 */
	public function duringLogin( &$user, $foundUser, &$returnPluginsOverrides ): void
	{
		global $_CB_database;

		$ipAddress					=	Application::Input()->getRequestIP();

		if ( ! CBAntiSpam::isBlockable( $user, $ipAddress ) ) {
			return;
		}

		$blocked					=	CBAntiSpam::getBlock( $user, $ipAddress );

		if ( $blocked ) {
			$this->blockLogin( $blocked );
			return;
		}

		if ( $this->params->getBool( 'login_duplicate', false ) ) {
			$sessions				=	CBAntiSpam::countUserSessions( $user->getInt( 'id', 0 ) );
			$count					=	$this->params->getInt( 'login_duplicate_count', 0 );

			if ( ! $count ) {
				$count				=	1;
			}

			if ( $sessions >= $count ) {
				$method				=	$this->params->getInt( 'login_duplicate_method', 1 );

				if ( $method > 0 ) {
					$reason			=	$this->params->getString( 'login_duplicate_reason', 'Already logged in.' );

					if ( $method > 1 ) {
						$row		=	new BlockTable();

						if ( $method === 2 ) {
							$row->setType( 'ip' );
							$row->setValue( $ipAddress );
						} elseif ( $method === 3 ) {
							$row->setType( 'account' );
							$row->setValue( $user->getInt( 'id', 0 ) );
						} else {
							$row->setType( 'user' );
							$row->setValue( $user->getInt( 'id', 0 ) );
						}

						$row->setDuration( $this->params->getString( 'login_duplicate_dur', '+1 HOUR' ) );
						$row->setReason( $reason );

						$row->store();

						$this->blockLogin( $row );
					} else {
						$this->blockLogin( $reason );
					}
				} else {
					$query			=	'DELETE'
									.	"\n FROM " . $_CB_database->NameQuote( '#__session' )
									.	"\n WHERE " . $_CB_database->NameQuote( 'userid' ) . " = " . $user->getInt( 'id', 0 );
					$_CB_database->setQuery( $query );
					$_CB_database->query();

					$query			=	'DELETE'
									.	"\n FROM " . $_CB_database->NameQuote( '#__user_keys' )
									.	"\n WHERE " . $_CB_database->NameQuote( 'user_id' ) . " = " . $_CB_database->Quote( $user->getString( 'username', '' ) );
					$_CB_database->setQuery( $query );
					$_CB_database->query();
				}
			}
		}

		if ( $this->params->getBool( 'login_share', false ) ) {
			$timeframe				=	$this->params->getString( 'login_share_timeframe', '-1 MONTH' );

			$query					=	'SELECT COUNT(*)'
									.	"\n FROM " . $_CB_database->NameQuote( '#__comprofiler_plugin_antispam_log' )
									.	"\n WHERE " . $_CB_database->NameQuote( 'user_id' ) . " = " . $user->getInt( 'id', 0 );
			if ( $timeframe ) {
				$query				.=	"\n AND " . $_CB_database->NameQuote( 'date' ) . " >= " . $_CB_database->Quote( Application::Date( 'now', 'UTC' )->modify( strtoupper( $timeframe ) )->format( 'Y-m-d H:i:s' ) );
			}
			$_CB_database->setQuery( $query );
			$logins					=	$_CB_database->loadResult();

			$count					=	$this->params->getInt( 'login_share_count', 10 );

			if ( ! $count ) {
				$count				=	10;
			}

			if ( $logins > $count ) {
				$method				=	$this->params->getInt( 'login_share_method', 0 );
				$reason				=	$this->params->getString( 'login_share_reason', 'Login sharing.' );

				if ( $method > 0 ) {
					$row			=	new BlockTable();

					if ( $method === 2 ) {
						$row->setType( 'ip' );
						$row->setValue( $ipAddress );
					} elseif ( $method === 3 ) {
						$row->setType( 'account' );
						$row->setValue( $user->getInt( 'id', 0 ) );
					} else {
						$row->setType( 'user' );
						$row->setValue( $user->getInt( 'id', 0 ) );
					}

					$row->setDuration( $this->params->getString( 'login_share_dur', '+1 HOUR' ) );
					$row->setReason( $reason );

					$row->store();

					$this->blockLogin( $row );
				} else {
					$this->blockLogin( $reason );
				}
			}
		}
	}

	/**
	 * Updates users ip address log and clears attempts
	 *
	 * @param UserTable $row
	 * @param bool      $loggedIn
	 * @param bool      $firstLogin
	 * @param array     $messagesToUser
	 * @param array     $alertMessages
	 * @param string    $return
	 */
	public function afterLogin( &$row, $loggedIn, $firstLogin, &$messagesToUser, &$alertMessages, &$return ): void
	{
		if ( ! $loggedIn ) {
			return;
		}

		CBAntiSpam::logIPAddress( $row, [ 'login', 'forgot' ] );
	}
}