<?php
/**
* Community Builder (TM)
* @version $Id: $
* @package CommunityBuilder
* @copyright (C) 2004-2023 www.joomlapolis.com / Lightning MultiCom SA - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

namespace CB\Plugin\ReconfirmEmail\Trigger;

use CBLib\Application\Application;
use CBLib\Registry\GetterInterface;
use CB\Database\Table\UserTable;
use CBLib\Language\CBTxt;
use CBLib\Input\Get;
use CB\Plugin\ReconfirmEmail\Helper;
use CB\Plugin\ReconfirmEmail\Table\ReconfirmEmailTable;
use CB\Database\Table\PluginTable;
use CB\Database\Table\FieldTable;

\defined( 'CBLIB' ) or die();

class UserTrigger extends \cbPluginHandler
{
	/**
	 * Deletes email changes when the user is deleted
	 *
	 * @param  UserTable $user
	 * @param  int       $status
	 */
	public function onAfterDeleteUser( $user, $status )
	{
		global $_CB_database;

		if ( ! $this->params->getBool( 'general_delete', true ) ) {
			return;
		}

		$query		=	'DELETE FROM ' . $_CB_database->NameQuote( '#__comprofiler_plugin_emails' )
					.	"\n WHERE " . $_CB_database->NameQuote( 'user_id' ) . " = " . $user->getInt( 'id', 0 );
		$_CB_database->setQuery( $query );
		$_CB_database->query();
	}

	/**
	 * @param string      $action
	 * @param string      $object
	 * @param string      $type
	 * @param UserTable   $user
	 * @param PluginTable $plugin
	 * @param FieldTable  $field
	 * @param mixed       $oldValues
	 * @param mixed       $newValues
	 * @param string      $reason
	 */
	public function onLogChange( $action, $object, $type, &$user, &$plugin, &$field, $oldValues, $newValues, $reason )
	{
		global $_PLUGINS;

		if ( ( $reason !== 'edit' ) || ( $field->getString( 'name', '' ) !== 'email' ) ) {
			// Skip if this isn't profile edit and isn't the email field:
			return;
		}

		if ( Application::MyUser()->isGlobalModerator() || Application::Application()->isClient( 'administrator' ) ) {
			// Moderators and backend do not require reconfirmation:
			return;
		}

		if ( $user->getError() || $_PLUGINS->getErrorMSG( false ) ) {
			// Validation of some kind has failed so we've nothing further to do:
			return;
		}

		$userId		=	$user->getInt( 'id', 0 );

		if ( Helper::getEmailChanged( $userId ) || Helper::getEmailChanging( $userId ) ) {
			// Skip since email address has already changed or it's currently being changed:
			return;
		}

		if ( ( $oldValues === $newValues ) || ( ! $oldValues ) || ( ! $newValues ) ) {
			// Nothing to change so don't bother with reconfirmation:
			return;
		}

		if ( Helper::emailExists( $newValues ) || ( ! cbIsValidEmail( $newValues ) ) ) {
			// Email already exists or isn't valid so skip notifying of email change and let Joomla user store error the profile update:
			return;
		}

		Helper::setEmailChanging( $userId, $newValues, $oldValues );
	}

	/**
	 * @param FieldTable  $field
	 * @param UserTable   $user
	 * @param PluginTable $postdata
	 * @param string      $reason
	 */
	public function onAfterPrepareFieldDataSave( &$field, &$user, &$postdata, $reason )
	{
		global $_PLUGINS;

		if ( ( $reason !== 'edit' ) || ( $field->getString( 'name', '' ) !== 'email' ) ) {
			// Skip if this isn't profile edit and isn't the email field:
			return;
		}

		if ( Application::MyUser()->isGlobalModerator() || Application::Application()->isClient( 'administrator' ) ) {
			// Moderators and backend do not require reconfirmation:
			return;
		}

		if ( $user->getError() || $_PLUGINS->getErrorMSG( false ) ) {
			// Validation of some kind has failed so we've nothing further to do:
			return;
		}

		$userId		=	$user->getInt( 'id', 0 );

		if ( ! Helper::getEmailChanging( $userId ) ) {
			// Skip since email address didn't change:
			return;
		}

		$newEmail	=	Helper::getNewEmail( $userId );

		if ( Helper::emailExists( $newEmail ) || ( ! cbIsValidEmail( $newEmail ) ) ) {
			// Email already exists or isn't valid so let normal Joomla email validation to fail profile update:
			Helper::setEmailChanging( $userId, null );

			return;
		}

		// Reset the email address:
		$user->set( 'email', Helper::getOldEmail( $userId ) );
	}

	/**
	 * @param UserTable $user
	 * @param UserTable $userDuplicate
	 * @param UserTable $oldUser
	 */
	public function onAfterUserUpdate( &$user, &$userDuplicate, $oldUser )
	{
		global $_CB_framework, $_CB_database;

		if ( Application::MyUser()->isGlobalModerator() || Application::Application()->isClient( 'administrator' ) ) {
			// Moderators and backend do not require reconfirmation:
			return;
		}

		$userId						=	$user->getInt( 'id', 0 );

		if ( Helper::getEmailChanged( $userId ) ) {
			// Skip since email address has already been changed:
			return;
		}

		if ( ! Helper::getEmailChanging( $userId ) ) {
			$newUserEmail			=	$user->getString( 'email', '' );
			$oldUserEmail			=	$oldUser->getString( 'email', '' );

			if ( $newUserEmail !== $oldUserEmail ) {
				// Our earlier behavior was interrupted so lets note the change now and reset below:
				Helper::setEmailChanging( $userId, $newUserEmail, $oldUserEmail );
			} else {
				// Skip since email address isn't being changed:
				return;
			}
		}

		$newEmail					=	Helper::getNewEmail( $userId );
		$oldEmail					=	Helper::getOldEmail( $userId );

		if ( ( $newEmail === $oldEmail ) || ( ! $newEmail ) || ( ! $oldEmail ) ) {
			// Nothing to change so don't bother with reconfirmation:
			Helper::setEmailChanging( $userId, null );

			return;
		}

		if ( $newEmail === $user->getString( 'email', '' ) ) {
			// Something has added the new email back into the user object so we need to force it to reset here:
			$user->storeDatabaseValue( 'email', $oldEmail, false );
			$user->set( 'email', $oldEmail );
		}

		$row						=	new ReconfirmEmailTable();

		$row->load( [ 'user_id' => $userId, 'to' => $newEmail, 'status' => 'P' ] );

		$message					=	CBTxt::T( $this->params->getHtml( 'reconfirm_changed', 'Your email address has changed and requires reconfirmation. Please check your new email address for your confirmation email.' ) );

		if ( $row->getInt( 'id', 0 ) ) {
			// Already have a pending reconfirm for this email address so just stop here:
			if ( $message ) {
				$_CB_framework->enqueueMessage( $message, 'message' );
			}

			Helper::setEmailChanging( $userId, null );

			return;
		}

		$cbUser						=	\CBuser::getInstance( $userId, false );

		// Cancel all other pending change requests as we only want 1 active change request:
		$query						=	'UPDATE ' . $_CB_database->NameQuote( '#__comprofiler_plugin_emails' )
									.	"\n SET " . $_CB_database->NameQuote( 'status' ) . " = " . $_CB_database->Quote( 'X' )
									.	"\n WHERE " . $_CB_database->NameQuote( 'user_id' ) . " = " . $userId
									.	"\n AND " . $_CB_database->NameQuote( 'status' ) . " = " . $_CB_database->Quote( 'P' );
		$_CB_database->setQuery( $query );
		$_CB_database->query();

		$row->set( 'user_id', $userId );
		$row->set( 'from', $oldEmail );
		$row->set( 'to', $newEmail );

		if ( $row->getError() || ( ! $row->check() ) ) {
			$_CB_framework->enqueueMessage( CBTxt::T( 'RECONFIRM_EMAIL_FAILED_TO_SAVE', 'Email address failed to save! Error: [error]', [ '[error]' => $row->getError() ] ), 'error' );

			return;
		}

		if ( $row->getError() || ( ! $row->store() ) ) {
			$_CB_framework->enqueueMessage( CBTxt::T( 'RECONFIRM_EMAIL_FAILED_TO_SAVE', 'Email address failed to save! Error: [error]', [ '[error]' => $row->getError() ] ), 'error' );

			return;
		}

		$reconfirm					=	$_CB_framework->pluginClassUrl( $this->element, true, [ 'confirmcode' => $row->getString( 'code' ) ] );
		$cancel						=	$_CB_framework->pluginClassUrl( $this->element, true, [ 'action' => 'cancel', 'confirmcode' => $row->getString( 'code' ) ] );

		$extra						=	[	'reconfirm'		=>	$reconfirm,
											'reconfirm_url'	=>	$reconfirm,
											'cancel'		=>	$cancel,
											'cancel_url'	=>	$cancel,
											'new_email'		=>	$newEmail,
											'old_email'		=>	$oldEmail,
											'user_url'		=>	$_CB_framework->viewUrl( 'userprofile', true, [ 'user' => $userId ] ),
										];

		$savedLanguage				=	CBTxt::setLanguage( $user->getUserLanguage() );

		$mailFromName				=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_from_name', '' ), true, false, $extra ), GetterInterface::STRING );
		$mailFromAddr				=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_from_address', '' ), true, false, $extra ), GetterInterface::STRING );
		$mailSubject				=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_subject', 'Your email address requires confirmation' ), true, false, $extra ), GetterInterface::STRING );
		$mailBody					=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_body', 'The email address attached to your account [username] has changed to [new_email] and requires confirmation.<br><br>You can confirm your email address by clicking on the following link:<br><a href="[reconfirm]">[reconfirm]</a><br><br>If this was done in error please contact administration or cancel by <a href="[cancel]">clicking here</a>.' ), false, false, $extra ), GetterInterface::HTML );
		$mailCC						=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_cc', '' ), true, false, $extra ), GetterInterface::STRING );
		$mailBCC					=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_bcc', '' ), true, false, $extra ), GetterInterface::STRING );
		$mailAttachments			=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_attachments', '' ), true, false, $extra ), GetterInterface::STRING );

		CBTxt::setLanguage( $savedLanguage );

		if ( $mailCC ) {
			$mailCC					=	preg_split( '/ *, */', $mailCC );
		}

		if ( $mailBCC ) {
			$mailBCC				=	preg_split( '/ *, */', $mailBCC );
		}

		if ( $mailAttachments ) {
			$mailAttachments		=	preg_split( '/ *, */', $mailAttachments );
		}

		$cbNotification				=	new \cbNotification();

		$cbNotification->sendFromSystem( $newEmail, $mailSubject, $mailBody, false, 1, $mailCC, $mailBCC, $mailAttachments, $extra, true, $mailFromName, $mailFromAddr );

		if ( $this->params->getBool( 'reconfirm_changed_notify', false ) ) {
			$mailSubject			=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_changed_subject', 'Your email address has changed' ), true, false, $extra ), GetterInterface::STRING );
			$mailBody				=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_changed_body', 'The email address attached to your account [username] has changed to [new_email] and requires confirmation.<br><br>Please check the new email address for a link to confirm this change. If this was done in error please contact administration or cancel by <a href="[cancel]">clicking here</a>.' ), false, false, $extra ), GetterInterface::HTML );
			$mailCC					=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_changed_cc', '' ), true, false, $extra ), GetterInterface::STRING );
			$mailBCC				=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_changed_bcc', '' ), true, false, $extra ), GetterInterface::STRING );
			$mailAttachments		=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_changed_attachments', '' ), true, false, $extra ), GetterInterface::STRING );

			if ( $mailCC ) {
				$mailCC				=	preg_split( '/ *, */', $mailCC );
			}

			if ( $mailBCC ) {
				$mailBCC			=	preg_split( '/ *, */', $mailBCC );
			}

			if ( $mailAttachments ) {
				$mailAttachments	=	preg_split( '/ *, */', $mailAttachments );
			}

			$cbNotification			=	new \cbNotification();

			$cbNotification->sendFromSystem( $oldEmail, $mailSubject, $mailBody, false, 1, $mailCC, $mailBCC, $mailAttachments, $extra, true, $mailFromName, $mailFromAddr );
		}

		if ( $this->params->getBool( 'reconfirm_moderators_notify', false ) ) {
			$mailSubject			=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_moderators_subject', '[username] email address has changed' ), true, false, $extra ), GetterInterface::STRING );
			$mailBody				=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_moderators_body', 'The email address attached to the account [username] has changed to [new_email] and is pending confirmation.' ), false, false, $extra ), GetterInterface::HTML );
			$mailCC					=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_moderators_cc', '' ), true, false, $extra ), GetterInterface::STRING );
			$mailBCC				=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_moderators_bcc', '' ), true, false, $extra ), GetterInterface::STRING );
			$mailAttachments		=	Get::clean( $cbUser->replaceUserVars( $this->params->getRaw( 'reconfirm_moderators_attachments', '' ), true, false, $extra ), GetterInterface::STRING );

			if ( $mailCC ) {
				$mailCC				=	preg_split( '/ *, */', $mailCC );
			}

			if ( $mailBCC ) {
				$mailBCC			=	preg_split( '/ *, */', $mailBCC );
			}

			if ( $mailAttachments ) {
				$mailAttachments	=	preg_split( '/ *, */', $mailAttachments );
			}

			$cbNotification			=	new \cbNotification();

			$cbNotification->sendToModerators( $mailSubject, $mailBody, false, 1, $mailCC, $mailBCC, $mailAttachments, $extra );
		}

		if ( $message ) {
			$_CB_framework->enqueueMessage( $message, 'message' );
		}

		Helper::setEmailChanging( $userId, null );
	}
}